/* checking.h                                                   (Gregor Meyer)
**                                                              (1991 - '92  )
**      c h e c k   a s s e r t i o n s
**      -------------------------------
**  Various macros to support debugging and writing of correct programs.
**  This is an enhanced version of <assert.h>.
**
**
**  Actions controlled by the macros:
**      *  check the integrity of an object (i.e. the invariant conditions),
**      *  check pre- and post-conditions of procedures/functions
**      *  check pre- and post-conditions and invariants of loops
**      *  check any other assertion
**      *  execute debug-statements
**      *  execute trace-statements
**
**  These macros are controlled by several preprocessor-flags in order
**  to switch the specific action on or off. The flags can be defined
**  locally, i.e. each time this file is included, in order to debug
**  or trace selected files.
*/


/*  Description:
**  -----------------------------------------------------------------
**  1)  check assertions
**      Note: 'B-expr' is a BOOLean expression
**
**      c h e c k ( B-expr )
**          Check if 'B-expr'is true at run-time.
**          If this check fails, the standard C-library function 'assert'
**          is used to abort the process with an appropriate message.
**
**      i f c h e c k ( stmt )
**          evaluate statement during run-time if check is enabled
**
**
**    The following macros are essentially the same as check(B-expr).
**    Special names are used to make their semantics explicit.
**    If you know the language Eiffel, you will notice severval similarities.
**
**      i n v a l i d ()               enforce check-failure
**
**      v a l i d ( B-expr)            a more 'declarative' notion of check
**
**      i n v a r i a n t ( B-expr )   an invariant-condition [of a loop]
**
**      r e q u i r e ( B-expr )       a pre-condition [of a loop or function]
**      precondition(b-expr)
**
**      e n s u r e ( B-expr )         a post-condition [of a loop or function]
**      postcondition(B-expr)
**
**      c h e c k _ i n v a r i a n t s ( func )
**      c l a s s _ i n v a r i a n t s ( B-func )
**          Call a BOOLean function 'B-func' to check if the [current] object
**          does not violate any invariant defined by the class.
**          Because class-invariants are not part of the C or C++ language,
**          a practical way to check these invariants is to create
**          a special function which test all these assertions. This function
**          may be called upon entry and/or exit of another function.
**
**  2)  Debugging support (if the system-debugger is not satisfactory)
**      Note: 'stmt'   is a statement
**
**      i f t r a c e ( stmt )
**          Evaluate statement during run-time if TRACE -flag is on.
**          Typically used as: iftrace(printf("now I'm here"));
**
** -----------------------------------------------------------------
**
**    F l a g s  to control previous macros by the pre-compiler
**
**          NOCHECK :  suppress all check-calls
**          NDEBUG  :    "          "         (this is the standard C flag)
**                      both flags are equivalent
**
**          CHECKFAIL_A
**          CHECKFAIL_T
**          OVERRIDE_ASSERTPROC:
**              Every check() is mapped onto Standard-C assert() by default.
**              If this mapping is overridden, the procedure 'checkfail' is
**              called (for a prototype see below), if check() fails at run-time.
**              This procedure must be defined somewhere by the user of this header.
**              The file checking.c can be used. CHECKFAIL_A will always abort,
**              CHECKFAIL_T calls the TROUBLE-function
**
**          TRACE   :  enable trace-calls
**
**      Defaults:
*/


#define CHECKING_H





/*------------------------------------------------------------------
**  private part, Implementation
*/


/* To be compatible with standard flag 'NDEBUG'. */
/* Make NDEBUG and NOCHECK equivalent.           */
#ifdef NDEBUG
#ifndef NOCHECK
#define NOCHECK
#endif
#endif
#ifdef NOCHECK
#ifndef NDEBUG
#define NDEBUG
#endif
#endif

#ifndef OVERRIDE_ASSERTPROC
#ifdef CHECKFAIL_T
#define OVERRIDE_ASSERTPROC
#endif
#ifdef CHECKFAIL_A
#define OVERRIDE_ASSERTPROC
#endif
#endif


#ifdef  NOCHECK    /* remove all checks  if no debug       */
#define checkdecl(decl)
#define ifcheck(stmt) ((void)0)
#define assert(b)   ((void)0) /* consistent with assert.h */
#define check(b)     ((void)0)
#define valid(b)     ((void)0)
#define invalid()    ((void)0)
#define require(b)   ((void)0)
#define ensure(b)    ((void)0)
#define invariant(b)   ((void)0)
#define precondition(b)  ((void)0)
#define postcondition(b)   ((void)0)
#define check_invariants(invproc)  ((void)0)

#else            /* CHECK enabled: */
#ifndef OVERRIDE_ASSERTPROC
#include <assert.h>
#define check(b)    assert(b)
#else
#define assert(b) check(b)
void  checkfail( char* cond, char * filename, int linenumber);
#ifndef SIMPLE_PREPROC
#define check(b)  ( (b) ? (void)0 : checkfail(#b,__FILE__,__LINE__) )
#else
#define check(b)  ( (b) ? (void)0 : checkfail("",__FILE__,__LINE__) )
#endif
/* #define check(b) {if(!(b)) checkfail(__FILE__,__LINE__);}             */
/*  error with:    if (xy) check(1); else check(2);   => misplaced else  */
#endif
#define checkdecl(decl)   decl
#define ifcheck(stmt)   {stmt;}
#define invalid()   check(FALSE)
#define valid(b)    check(b)
#define require(b)    check(b)
#define ensure(b)   check(b)
#define invariant(b)  check(b)
#define precondition(b) check(b)
#define postcondition(b) check(b)
#define check_invariants(invproc)    invproc
#define class_invariants(bfunc)    check(bfunc)
#endif  /* NOCHECK else */



#ifdef TRACE
#define iftrace(stmt)   {stmt; fflush(stdout); fflush(stderr);}
#else
#define iftrace(stmt)   ((void)0)
#endif




/* Play with several levels/priorities of checking */
/* This may be too sophisticated                   */
#define assert0(b)  check(b)     /*  0: highest priority */
#define assert1(b)  check(b)
#define assert2(b)  check(b)
#define assert3(b)  check(b)
#ifndef NDEBUG9
#define assert9(b)  check(b)
#else
#define assert9(b)
#endif


