%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(typeops).
:- ensure_loaded(glib).
:- ensure_loaded(csp_h).
:- ensure_loaded(csppo). % this is for the clauses

:- type_import(csppo).  % this is for the type declarations
:- type_import(glib).
:- type_import(checking).
:- type_import(prolog_types).
:- type_import(prolog_meta).
:- type_import(prolog_num).
:- type_import(prolog_io).
:- type_import(prolog_misc).
:- type_import(csp_h).
  
type poset(T) --> poset(list(T),list(order1(T))).
type null --> null_elem.

% public functions

% po_create( +Domain, +Order, -PO) :-
%       PO is the internal representation of partial Order over Domain.
%       Domain plus parameter symbols A and Order are inserted into C-Programm
% 

po_create(Dom,Ord,PO) :-
	c_free_space,
	c_set_null_elem(null_elem),
	(check_cycle(Ord,Cycle) -> 
	    (write('cycle in type declarations: '),write(Cycle),write(Ord), nl, halt)
	    ;
	    (allowed_parameter_symbols(A),
	    append(Dom,A,DomA),
	    remove_duplicates(DomA,DomB),
	    real_po_create(DomB,Ord,PO))).

% po_extend( +PO, +Ids, -POnew) :-
%       Ids are added to domain of PO, result is POnew.
%       C implementation does not handle dynamic extension of posets, 
%       only checks if Ids are already stored in PO
%

po_extend(PO,L,POe) :-
	check_allowed(L),
	POe = PO.

% satisfy_constraints_once(+PO, +-CONSTRAINTS) :-
%       solves the CSP problem described by POSET and CONSTRAINTS, 
%       variables in CONSTRAINTS are bound to one set of values 
%       that satisfy the CSP
% 

satisfy_constraints_once(_PO,Constraints) :-
%write(satisfy),nl,write(Constraints),nl,
        c_reset_constraints,
        insert_constraints(Constraints,VB),!,
	c_solvable_first(1),!,
        enumerate_solutions(VB,S),
        bind(S)
%,write(Constraints),nl
.

% satisfy_constraints(+PO, +-CONSTRAINTS) :-
%       solves the CSP problem described by POSET and CONSTRAINTS, 
%       variables in CONSTRAINTS are bound to values to satisfy the CSP
% 

satisfy_constraints(_PO,Constraints) :-
        c_reset_constraints,!,
        insert_constraints(Constraints,VB),!,
	c_solvable_first(1),!,
        c_solvable_all(1),!,
        enumerate_solutions(VB,S),
        bind(S).

% po_test_leq( +PO, +X, +Y) :- 
%	(X =< Y) is true in PO.
%
po_test_leq(_,X,Y) :-
	c_test_leq(X,Y,R),
	!,R = 1.

% po_lub(PO,X,Y,Z) :- 
% 	Z is least upper bound of X and Y in partial order PO,
%	fails if no upper bound exists or lub is not unique.
%
po_lub(_,X,Y,Z) :-
	c_lub(X,Y,Z),
	Z \== null_elem .

%----
% po_maximal( + PO, + X) :-
%       Element X is maximal in the order of PO
%
po_maximal(_,X) :-
	c_is_maximal(X,R),
	!,R = 1.
 
 
%----
% po_minimal( + PO, + X) :-
%       Element X is minimal in the order of PO
%
po_minimal(_,X) :-
	c_is_minimal(X,R),
	!,R = 1.
 
 
%----
% po_glb( +PO, +X, +Y, ?Z) :-
%       Z is greatest lower bound of X and Y in partial order PO,
%       fails if no lower bound exists or glb is not unique.
%
po_glb(_,X,Y,Z) :-
	c_glb(X,Y,Z),
	Z \== null_elem.


%----
% po_write(PO) :- 
%       writes out the domain and order relations of PO
%
po_write(poset(D,O)) :-
        write(poset(D,O)).  % don't write tables

 
%---- 
% po_indomain(PO,X) :- 
%       Element X is element in the domain of PO 
% 
po_indomain(poset(D,_O),X) :- 
	member(X,D).
 
 
%---- 
% po_ordered(PO,X) :- 
%       Element X is element in the order of PO 
% 
po_ordered(poset(_,O),X) :- memberchk(X=<_Y,O),!,true.
po_ordered(poset(_,O),X) :- memberchk(_Y=<X,O), true.
 
 
%----
% po_less(PO,X,Y) :- (X =< Y) and X \= Y is true in PO.
%
po_less(_,X,Y) :- 
	ground(X),
	ground(Y),
	c_test_leq(X,Y,1),
	\+ ( X = Y).
po_less(_,X,Y) :-
	write('error: po_less: cannot work on non ground terms '),
	write(X),write(' and '), write(Y),nl,abort.

% local functions

local pred allowed_parameter_symbols(list(top)).
allowed_parameter_symbols(['#1','#2','#3','#4','#5','#6','#7']).
% see decl.pl: my_number_tvars

local pred check_allowed(list(_)).
check_allowed(Used) :-
	allowed_parameter_symbols(A),
	subset(Used,A),
	!.
check_allowed(Used) :-
	write('error: internal limitation'),nl,
	write(Used), 
	write('is not subset of'),nl,
	allowed_parameter_symbols(A),
	write(A),nl,
	fail.



% real_po_create(Domain,Order,PO) :-
%       PO is the internal representation of partial Order over Domain.
%       Domain plus parameter symbols A and Order are inserted into C-Programm

local pred real_po_create(list(T),list(order1(T)),poset(@T)).

real_po_create(Domain,Order,poset(Domain,Order)) :-
        length(Domain,N1),!,
        c_constant_num(N1),
        insert_constants(Domain),
        length(Order,N2),!,
        c_order_num(N2),
        insert_order_rel(Order),!,
	c_check_unique(_N),!.
 

%-------
% insert_constants(Constants) :-
%       calls for each atom in list Constants the C-function constant().
%
local pred insert_constants(list(_T)).
%
insert_constants([]).
insert_constants([C|Rest]) :- c_constant(C),insert_constants(Rest).
 
%-------
% insert_order_rel(Ordering) :-
%       calls for each order relation in list Ordering the C-function
%       order_def_less().
%
 
local pred insert_order_rel(list(order1(_T))).
 
insert_order_rel([]).
insert_order_rel([C1 =< C2 | Rest]) :-
        c_order_def_less(C1, C2),insert_order_rel(Rest).
 
%-------
% insert_constraints(Constraints,VariableBindings) :-
%	the constraints in Constraints are counted and sorted in binary
%       and upper bound constraints. The occuring variables are counted
%	and each one gets an unique atom assigned. The variables and the
%       constraints are inserted into the C-Constraint solver.
%

local pred insert_constraints(list(po_constraint(_T)),list(top)).
 
insert_constraints(Cons,VB) :-
        count_constraints(Cons,ConsBC,NumBC,ConsUB,NumUB,Terme),
        var_remove_duplicates(Terme,TermeF),
        count_variables(TermeF,VB,NumV), 
        c_variable_num(NumV),
        insert_variables(VB),
        c_binary_constraint_num(NumBC),
        insert_binary(VB,ConsBC),
        c_upper_bound_constraint_num(NumUB),
        insert_upper_bound(VB,ConsUB).

 

local pred count_constraints(list(po_constraint(T)),list(po_constraint(T)),nat,list(po_constraint(T)),nat,list(T)).
 
count_constraints([],[],0,[],0,[]).
count_constraints([ X =< Y | Rest],[ X =< Y | RestBC],NumBC,ConsUB,NumUB,[X,Y|RestVar]) :-
	count_constraints(Rest,RestBC,N,ConsUB,NumUB,RestVar),
        NumBC is N +1.
count_constraints([ lub(X,Y) | Rest],ConsBC,NumBC,[ lub(X,Y) | RestUB],NumUB,Var) :-
        append(Y,[X|Restvar],Var),
        count_constraints(Rest,ConsBC,NumBC,RestUB,N,Restvar),
        NumUB is N +1.
 
local pred count_variables(list(_T),list(top),nat).
 
count_variables([],[],0).
count_variables([T|RestT],[[T,Atom]|RestVB],Num) :-
        var(T),
	count_variables(RestT,RestVB,N),
	Num is N + 1,
        number_chars(Num,L),
	atom_chars(Atom,[120,120,120|L]).
count_variables([T|RestT],VB,Num) :-
        nonvar(T),
	count_variables(RestT,VB,Num).
 
local pred insert_variables(list(top)).
 
insert_variables([]).
insert_variables([[_,N]|RestV]) :-
        c_variable(N),
	insert_variables(RestV).
 
local pred insert_binary(list(top),list(po_constraint(_T))).
 
insert_binary(_,[]).
insert_binary(VB,[X =< Y | Rest]) :-
        var(X),var(Y),find_var(VB,X,V1),find_var(VB,Y,V2),
        c_constraint_def_vv(V1,V2),insert_binary(VB,Rest).
insert_binary(VB,[X =< Y | Rest]) :-
        nonvar(X),var(Y),find_var(VB,Y,V1),
        c_constraint_def_ev(X,V1),insert_binary(VB,Rest).
insert_binary(VB,[X =< Y | Rest]) :-
        var(X),nonvar(Y),find_var(VB,X,V1),
        c_constraint_def_ve(V1,Y),insert_binary(VB,Rest).
insert_binary(VB,[X =< Y | Rest]) :-
        nonvar(X),nonvar(Y),c_constraint_def_ee(X,Y),insert_binary(VB,Rest).
 
local pred insert_upper_bound(list(top),list(po_constraint(_T))).
 
insert_upper_bound(_,[]).
insert_upper_bound(VB,[lub(X,[A|Y])|Rest]) :-
        var(A),
	find_var(VB,A,V),
        c_ub_first_v(V,Expr),
	do_list(VB,Expr,Y,E),
	do_head(VB,E,X),
        insert_upper_bound(VB,Rest).
insert_upper_bound(VB,[lub(X,[A|Y])|Rest]) :-
        nonvar(A),
	c_ub_first_e(A,Expr),
	do_list(VB,Expr,Y,E),
	do_head(VB,E,X),
        insert_upper_bound(VB,Rest).
 
local pred do_list(list(top),top,list(top),top).
 
do_list(_,Expr,[],Expr).
do_list(VB,Expr,[T|Rest],E) :-
        var(T),
	find_var(VB,T,V),
        c_ub_expr_v(Expr,V,E1),
	do_list(VB,E1,Rest,E).
do_list(VB,Expr,[T|Rest],E) :-
        nonvar(T),
	c_ub_expr_e(Expr,T,E1),
	do_list(VB,E1,Rest,E).
 
local pred do_head(list(top),top,top).
 
do_head(VB,E,V) :- 
	var(V),
	find_var(VB,V,V1),
	c_constraint_def_vu(V1,E).
do_head(_,E,C) :- 
	nonvar(C),
	c_constraint_def_eu(C,E).

local pred find_var(list(top),top,top).  
find_var([[V1,A]|_],V,A) :- V == V1. 
find_var([_|VB],V,A) :- find_var(VB,V,A).
 
local pred find(list(top),top,top).
find([[V1,A]|_],V,A) :- V = V1. 
find([_|VB],V,A) :- find(VB,V,A).
 
local pred enumerate_solutions(list(top),top).
enumerate_solutions(VB,S) :- 
	c_get_count_solution(CN),
	CN =\= 0,               
	enumerate(VB,0,CN,S).
 
local pred enumerate(list(top),nat,nat,top).
enumerate(VB,I,_,S) :- 
	c_get_count_variable(NV),
	tupel(VB,I,0,NV,S).
enumerate(VB,I,CN,S) :- 
	I2 is I + 1, I2 < CN, 
	enumerate(VB,I2,CN,S).
 
local pred tupel(list(top),nat,nat,nat,list(top)).
tupel(_,_,J,NV,[]) :- J >= NV,!.
tupel(VB,I,J,NV,[[VAR,S1]|S]) :- 
	c_get_variables(J,V),
	find(VB,VAR,V),
        c_get_solutions(I,J,S1),
        J1 is J + 1,
	tupel(VB,I,J1,NV,S).
 
local pred bind(list(top)).
bind([]).
bind([[V,C]|Rest]) :-
        (var(V) -> ( V = C) ; true),
        bind(Rest).
 
 


































































