%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(typeops).

:- ensure_loaded(glib).
:- ensure_loaded(csppo). % this is for the clauses

:- type_import(csppo).  % this is for the type declarations
:- type_import(prolog_types).
:- type_import(prolog_io).
:- type_import(glib).
:- type_import(checking).

%----------------------------------------------

% A partial order over some domain D has a binary relation D =< D
% where =< is reflexive, transitive and antisymmetric.
%
% A quasi-lattice is a partial (D, =<), such that
% + if two elements have an upper bound, then there is a 
%	unique minimal upper bound, called least upper bound (lub)
% + if two elements have a  lower bound, then there is a 
%	unique lower upper bound, called greatest lower bound (glb)
%
% A upper semi-lattice is a quasi-lattice, such that
% + lub's always exist (and are unique by definition)
% A lower semi-lattice ... glb's exist
%
% A lattice is an upper and lower semi-lattice
% i.e. lub's and glb's always exist


:- local type mub3(T) --> mub3(T,T,T).

:- local type p0set(T) --> 
	p0set( list(T),            % domain
	       list(order1(T))     % basic order relationships
	     ).

% A partially ordered set is internally represented by a structure
:- abstract type poset(T) --> 
	poset( p0set(T),           % domain + order
	       list(order1(T)),    % tabulation of all X =< Y, (X\=Y)
	       list(mub3(T))       % tabulation of all X \mub Y = Z, (X\=Y)
	     ).

%:- type qlat(T) == poset(T).
% should be 
%     :- subtype qlat(T) < poset(T).


:- type special_type --> bottom;top.


% internal naming convention:
% po...   works on poset
% p0...   works on p0set

% functions defined in csppo.pl :

%----
% po_create(Domain,Order,PO) :-
%	PO is the internal repr. of partial order over Domain.
%
po_create(Domain,Order,PO) :-
%    nl,write(creating_poset(Domain,Order)),nl,write('....'),flush_output(user),
	(check_cycle(Order,Cycle) -> 
            (write('cycle in type declarations: '),
	     write(Cycle),write(Order), nl, abort)
            ;	
	    (remove_duplicates(Domain,D),
	     normalize_p0set_ord(Order,O),
	     leq_table(p0set(D,O),LT),
	     % alternatives
	     % mub_table_naive(p0set(D,O),MT),
	     mub_table_it(D,LT,MT),
%	     write(ok),nl,
	     PO = poset(p0set(D,O),LT,MT))).

%test po_create([a,b,c],[a=<c],X).
%     X=poset([a,b,c],[a=<c],[a=<c],[mub3(a,c,c),mub3(c,a,c)]) 


%----
% po_extend(PO, Ids, POnew) :-
%	Ids are added to domain of PO,
%	result is POnew.
%
po_extend(poset(p0set(D,O),LT,MT),Ids,poset(p0set(Dnew,O),LT,MT)) :-
	append(D,Ids,Dnew).


%----
% po_test_leq(PO,X,Y) :- (X =< Y) is true in PO.
%
po_test_leq(_PO,X,X).
po_test_leq(poset(_,LT,_),X,Y) :- member(X=<Y,LT).


%----
% po_lub(PO,X,Y,Z) :- 
% 	Z is least upper bound of X and Y in partial order PO,
%	fails if no upper bound exists or lub is not unique.
%
po_lub(PO,X,Y,Z) :-
	po_mub(PO,X,Y,Z),
	\+ (po_mub(PO,X,Y,Z2), \+ Z = Z2).


%---- 
% po_maximal(PO,X) :- 
%       Element X is maximal in the order of PO 
% 
po_maximal(poset(p0set(_,O),_,_),X) :- \+ (memberchk(X=<_Y,O),true).


%---- 
% po_minimal(PO,X) :- 
%       Element X is minimal in the order of PO 
% 
po_minimal(poset(p0set(_,O),_,_),X) :- \+ (memberchk(_Y=<X,O), true).

% some more functions on posets 

%----
% po_write(PO) :- 
%       writes out the domain and order relations of PO
%
po_write(poset(p0set(D,O),_,_)) :-
	write(poset(D,O,'$lt','$mt')).  % don't write tables

%---- 
% po_indomain(PO,X) :- 
%       Element X is element in the domain of PO 
% 
po_indomain(poset(p0set(D,_O),_,_),X) :- member(X,D).


%---- 
% po_ordered(PO,X) :- 
%       Element X is element in the order of PO 
% 
po_ordered(poset(p0set(_,O),_,_),X) :- memberchk(X=<_Y,O),!,true.
po_ordered(poset(p0set(_,O),_,_),X) :- memberchk(_Y=<X,O), true.


%----
% po_less(PO,X,Y) :- (X =< Y) and X \= Y is true in PO.
%
po_less(poset(_,LT,_),X,Y) :- member(X=<Y,LT), true.


%----
% po_mub(PO,X,Y,Z) :- Z is minimal upper bound of X and Y.
%
po_mub(_PO,X,X,X).
po_mub(poset(_,_,MT),X,Y,Z) :- member(mub3(X,Y,Z),MT).
po_mub(poset(_,_,MT),X,Y,Z) :- member(mub3(Y,X,Z),MT).


%----
% po_glb(PO,X,Y,Z) :- 
% 	Z is greatest lower bound of X and Y in partial order PO,
%	fails if no lower bound exists or glb is not unique.
%
po_glb(poset(P0,_,_),X,Y,Z) :-
	p0_glb(P0,X,Y,Z).


% local functions

%----
% leq_table
% ersetzen durch setof-Aufruf, sonst zu ineffizient
%   setof(X=<Y,(po_0_leq(D,O,X,Y),\+X=Y),LT).
%
:- local pred leq_table(p0set(@T),list(order1(T))).
%
leq_table(P0,LT) :-
	leq_table(P0,[],LT).
%   setof(X=<Y,(p0_less(P0,X,Y)),LT).

:- local pred leq_table(p0set(@T),list(order1(T)),list(order1(T))).
leq_table(P0,LT0,LT) :-
	p0_less(P0,X,Y),   true,
	true,
	\+ memberchk(X=<Y,LT0),
	!,
	leq_table(P0,[X=<Y|LT0],LT).
leq_table(_,LT,LT).

:- local pred mub_table_naive(p0set(@T),list(mub3(T))).
mub_table_naive(P0,MT) :-
	mub_table_naive(P0,[],MT).

:- local pred mub_table_naive(p0set(@T),list(mub3(T)),list(mub3(T))).
mub_table_naive(P0,MT0,MT) :-
	p0_mub(P0,X,Y,Z),   true,
	\+ X=Y,
	X @< Y,   % mub is symmetric
	\+ member(mub3(X,Y,Z),MT0),
	!,
	mub_table_naive(P0,[mub3(X,Y,Z)|MT0],MT).
mub_table_naive(_,MT,MT).


:- local pred p0_indomain(p0set(@T),T).
p0_indomain(p0set(D,_O),X) :- member(X,D).

:- local pred p0_leq_base(p0set(@T),T,T).
p0_leq_base(p0set(_D,O),X,Y) :- member( X=<Y, O).

:- local pred check_p0set(p0set(_)).
check_p0set(_PO) :-
	true,
	true,
	true,
	true,
	true,
	true,
	% true,
	true.  % no redundancy in base.
	% true.


:- local pred p0_leq(p0set(@T),T,T).
%p0_leq(PO,_,_) :- check_p0set(PO), fail.
p0_leq(PO, X,X) :- p0_indomain(PO,X).        % reflexive
p0_leq(PO, X,Y) :- p0_leq2(PO,X,Y).
p0_leq(PO,bottom,Y) :- type_pragma(non_generic),
	p0_indomain(PO,bottom),
	p0_indomain(PO,Y). 
p0_leq(PO,X,top) :- type_pragma(non_generic),
	p0_indomain(PO,top),
	p0_indomain(PO,X). 

:- local pred p0_leq2(p0set(@T),T,T).
p0_leq2(PO, X,Y) :- 
	p0_leq_base(PO, X,Y).
p0_leq2(PO, X,Y) :- 
	p0_leq_base(PO, X,Z), 
	p0_leq2(PO, Z,Y).  % transitive


/*  efficiency */
:- local pred p0_less(p0set(@T),T,T).
% p0_less(PO, X,Y) :- p0_leq(PO,X,Y), \+ X=Y.
p0_less(PO,bottom,Y) :-  type_pragma(non_generic), 
	p0_indomain(PO,bottom),
	p0_indomain(PO,Y), 
	\+ Y = bottom.
p0_less(PO,X,top) :- type_pragma(non_generic), 
	p0_indomain(PO,top),
	p0_indomain(PO,X), 
	\+ X=top.
p0_less(PO, X,Y) :-
	p0_leq_base(PO, X,Y), true.
p0_less(PO, X,Y) :-
	p0_leq_base(PO, X,Z),
	p0_less(PO, Z,Y).  % transitive

:- local predl p0_minimal(p0set(T),T).
p0_minimal(PO,X) :- %  implicit with predl: type_pragma(non_generic), 
	( X = bottom 
	; \+ p0_less(PO,_,X)
	).

% auch x<Y in base erlauben ??
% bottom < #i-of-head true ?
% x=<y,y<z => x<z
% x<y,y=<z => x<z


% p0_is_antisymmetric(PO) :- \forall x,y if x=<y,y=<x then x=y
:- local pred p0_is_antisymmetric(p0set(_)).
p0_is_antisymmetric(PO) :- 
	\+ (p0_leq(PO,X,Y), \+ X = Y, p0_leq(PO,Y,X)).
	%% \+ (p0_leq(PO,X,Y), p0_leq(PO,Y,X), \+ X=Y). infinite loop



:- type order_table(T) == list(order1(T)).
% an order table is expected to contain
% the irreflexive transitive closure of some partial order
% synonym: leq-table


 
%---- 
% p0_max_member(LT,X,S) :-
%       relative to ordering given by LT,
%       X is a maximal member of set S.
%
:- local pred p0_max_member(order_table(@T),T,list(T)).
%
p0_max_member(LT,X,S) :-
        member(X,S),
        \+ ( member(X=<Y,LT),   true,
             member(Y,S)
           ).

:- local pred p0_max_member_elim(order_table(@T),T,list(T),list(T)).
%
p0_max_member_elim(LT,X,S,S0) :-
        member_elim(X,S,S0),
        \+ ( member(X=<Y,LT),   true,
             member(Y,S)
           ).
 
 
%---- 
% p0_add_lb(LT,S,X,SX) :-
%       relative to ordering given by LT
%       and a set S of minimal elements
%       X is added to S if it is minimal w.r.t S,
%
:- local pred p0_add_lb(order_table(@T),list(T),T,list(T)).
%
p0_add_lb(LT,S,X,SX) :-
        member(Y,S),
        member(Y=<X,LT),
        !,
        SX = S.
p0_add_lb(LT,S,X,SX) :-
        member_elim(Y,S,S0),
        member(X=<Y,LT),
        !,
        p0_add_lb(LT,S0,X,SX).
p0_add_lb(_LT,S,X,SX) :-
        % \+ X=<Y, \+ Y =<X
        SX = [X|S].
 
 


:- local pred p0_glb(p0set(@T),T,T,T).  
% Z = greatest lower bound of X,Y
p0_glb(PO, X,Y,Z) :- p0_leq(PO, Z,X), p0_leq(PO, Z,Y), true,
              \+ ( p0_leq(PO, ZZ,X), p0_leq(PO, ZZ,Y), \+ p0_leq(PO, ZZ,Z) ).

:- local pred p0_mlb(p0set(@T),T,T,T).  
% Z = maximal lower bound of X,Y
p0_mlb(PO, X,Y,Z) :- p0_leq(PO, Z,X), p0_leq(PO, Z,Y), true,
              \+ ( p0_leq(PO, ZZ,X), \+ Z=ZZ, p0_leq(PO, ZZ,Y), 
		   p0_leq(PO, Z,ZZ) ).

:- local pred p0_lub(p0set(@T),T,T,T).
% p0_lub(PO, X,Y) = Z         least upper bound
p0_lub(PO, X,Y,Z) :- p0_leq(PO, X,Z), p0_leq(PO, Y,Z), true,
              \+ ( p0_leq(PO, X,ZZ), p0_leq(PO, Y,ZZ), \+ p0_leq(PO, Z,ZZ) ).

:- local pred p0_mub(p0set(@T),T,T,T).
% p0_mub(PO, X,Y) = Z      minimal upper bound
p0_mub(PO, X,Y,Z) :- p0_leq(PO, X,Z), p0_leq(PO, Y,Z), true,
              \+ ( p0_leq(PO, X,ZZ), \+ Z=ZZ, p0_leq(PO, Y,ZZ), 
		   p0_leq(PO, ZZ,Z) ).

:- local pred lubt_list(poset(@T),list(T),T).
% lubt_list(Lat, L,Z) :- Z is lub of set/list L
lubt_list(_,[],_) :- write('!? lub of empty set ??'), abort.
lubt_list(_Lat, [X],X).
lubt_list(Lat, [X,Y|L],Z) :-
	po_lub(Lat, X,Y,Z0),
	lubt_list(Lat, [Z0|L],Z).


:- local pred normalize_p0set_ord(list(top),list(top)).
%
normalize_p0set_ord([],[]) :- !.
normalize_p0set_ord([X=<Y|C],O) :-
	true,
	X=Y, !,
	normalize_p0set_ord(C,O).
/* no no
normalize_p0set_ord([X=<Y|C],[X<Y|O]) :- 
	check(ground(X:Y)),
	\+ X=Y,   !,
	normalize_p0set_ord(C,O).
*/
normalize_p0set_ord([X=<Y|C],[X=<Y|O]) :- !,
	normalize_p0set_ord(C,O).
normalize_p0set_ord(_A,_B) :- true.



%-------------------------------------------------
%       given an order table LT
%       construct a table of mub-triples
:- local pred mub_table_it(list(T),list(order1(T)),list(mub3(T))).
%
mub_table_it(Domain,LT,MT) :-           % iterative construction
        mub_tab(LT,[],Domain,[],MT).

% mub_tab([a=<b,c=<b,c=<d],[],[a,b,c,d],[],X).
% mub_tab(LT,Done,S,T0,T) :- write(mub_tab(LT,Done,S,T0,T)),nl,fail.
:- local pred mub_tab(list(order1(T)),list(T), list(T),list(mub3(T)),list(mub3(T))).
%
mub_tab(_LT,_Done,[],T0,T0).
mub_tab(LT,Done,S,T0,T) :-
        p0_max_member_elim(LT,X,S,S0),
        !,
        % for all members Y of Done (i.e. complement_S)
        p0_mt(Done,LT,X,T0,T1),
        mub_tab(LT,[X|Done],S0,T1,T).

% X is maximal w.r.t. complement of Done
% p0_mt(Done,_LT,X,T0,Tab) :- write(p0_mt(Done,lt,X,T0,Tab)),nl,fail.
:- local pred p0_mt(list(T),list(order1(T)),T,list(mub3(T)),list(mub3(T))).
%
p0_mt([],_LT,_X,T0,T0).
p0_mt([Y|Done],LT,X,T0,Tab) :-
        p0_up_neighbors(X,LT,U),  % write('  '),write('U'=U),nl,
        p0_mt_mub_set(U,Y,T0,MU), % write('  '),write('MU'=MU),nl,
	% MU is set of B with (B = U1 \mub Y, U1 in U)
        p0_mt_minb(MU,LT,[],S),   % write('  '),write('S'=S),nl,
        p0_mt_add_mubs(T0,X,Y,S,T2), % write('  '),write('T2'=T2),nl,
	!,
        p0_mt(Done,LT,X,T2,Tab).

        
% find minimal bounds of U
% p0_mt_minb(U,_LT,Sa,S) :- write(p0_mt_minb(U,lt,Sa,S)),nl,fail.
:- local pred p0_mt_minb(list(T),order_table(T),list(T),list(T)).
%
p0_mt_minb([],_LT,S,S).
p0_mt_minb([B|U],LT,Sa,S) :-
        p0_add_lb(LT,Sa,B,SB),
        p0_mt_minb(U,LT,SB,S).

        
% p0_mt_add_mubs(T0,X,Y,Bs,T) :-
%       Bs is list of minimal upper bounds of X,Y
%       T is T0 extended by mub-entries  mub3(X,Y,B)
% p0_mt_add_mubs(T0,X,Y,Bs,T) :- write(p0_mt_add_mubs(T0,X,Y,Bs,T)),nl,fail.
:- local pred p0_mt_add_mubs(list(mub3(T)),T,T,list(T),list(mub3(@T))).
%
p0_mt_add_mubs(T0,_X,_Y,[],T0).
p0_mt_add_mubs(T0,X,Y,[B|Bs],T) :-
        p0_mt_mub3(X,Y,B,MUB),
        p0_mt_add_mubs([MUB|T0],X,Y,Bs,T).


:- local pred p0_mt_mub3(T,T,T,mub3(@T)).
%
p0_mt_mub3(X,Y,Z,mub3(X,Y,Z)) :- X @< Y.
p0_mt_mub3(X,Y,Z,mub3(Y,X,Z)) :- X @> Y.

:- local pred p0_mt_mub(list(mub3(@T)),T,T,T).
%
p0_mt_mub(_,X,X,X).
p0_mt_mub(MT,X,Y,Z) :- member(mub3(X,Y,Z),MT).
p0_mt_mub(MT,X,Y,Z) :- member(mub3(Y,X,Z),MT).



%----
% p0_up_neighbors(X,LT,U) :-
%       U is set of Y with X=<Y in LT
%
:- local pred p0_up_neighbors(T,order_table(@T),list(T)).
%
p0_up_neighbors(_X,[],[]).
p0_up_neighbors(X,[X=<Y|LT],[Y|U]) :-
        !,
        p0_up_neighbors(X,LT,U).
p0_up_neighbors(X,[_|LT],U) :-
        p0_up_neighbors(X,LT,U).

% mub_set(S,Y,MT,U) :-
:- local pred p0_mt_mub_set( list(T),T,list(mub3(@T)),list(T)).
p0_mt_mub_set([],_,_,[]).
p0_mt_mub_set([X|S],Y,MT,[B|U]) :-
        p0_mt_mub(MT,X,Y,B),
	!,
        p0_mt_mub_set(S,Y,MT,U).
p0_mt_mub_set([_|S],Y,MT,U) :-
        p0_mt_mub_set(S,Y,MT,U).






%----
% extend_poset(PO, Ids, Order, POnew) :-
%       Ids are added to domain of PO,
%       order of PO is extended with Order,
%       result is POnew.
%
:- local pred extend_posetC(poset(T),list(T),list(order1(T)),poset(@T)).
%pred extend_posetC(poset(T),list(T),list(po_constr(T)),poset(@T)).
%
extend_posetC(poset(p0set(D,O),LT,MT),Ids,C,poset(p0set(Dnew,O),LT,MT)) :-
        % true,
        C = [],
        !,
        append(D,Ids,Dnew).
extend_posetC(PO,Ids,C,_) :-
        write(poset:PO),nl,
        write(Ids:C:'can not extend poset, implementation missing'),
        abort. 

/*
complete_bottom(poset(T,O),poset(T,O)) :- member(bottom,T), !.
complete_bottom(poset(T,O),poset([bottom|T],O)).
complete_top(poset(T,O),poset(T,O)) :- member(top,T), !.
complete_top(poset(T,O),poset([top|T],O)).
*/

/*
complete_bottom(poset(T,O),poset(Tb,Ob)) :-
        (member(bottom,T) -> Tb = T
        ; Tb = [bottom|T]
        ),
        make_less_bottom_pairs(T,BT),
        append(BT,O,Ob0),  check(ground(Ob0)),
        remove_duplicates(Ob0,Ob).

make_less_bottom_pairs([],[]).
make_less_bottom_pairs([Y/_|Ys],[bottom<Y|XYs]) :-  !,  check( \+ Y=bottom),
        make_less_bottom_pairs(Ys,XYs).
make_less_bottom_pairs([Y|Ys],[bottom<Y|XYs]) :-  \+ Y=bottom,
        make_less_bottom_pairs(Ys,XYs).

make_leq_bottom_pairs([],[]).
make_leq_bottom_pairs([Y/_|Ys],[bottom=<Y|XYs]) :-  !,
        make_leq_bottom_pairs(Ys,XYs).
make_leq_bottom_pairs([Y|Ys],[bottom=<Y|XYs]) :-
        make_leq_bottom_pairs(Ys,XYs).


complete_top(poset(T,O),poset(Tt,Ot)) :-
        Tt = [top|T],
        make_less_top_pairs(T,BT),
        append(BT,O,Ot).

make_less_top_pairs([],[]).
make_less_top_pairs([Y/_|Ys],[Y<top|XYs]) :- !,  check( \+ Y=top),
        make_less_top_pairs(Ys,XYs).
make_less_top_pairs([Y|Ys],[Y<top|XYs]) :-
        make_less_top_pairs(Ys,XYs).

make_leq_top_pairs([],[]).
make_leq_top_pairs([Y/_|Ys],[Y=<top|XYs]) :- !,  check( \+ Y=top),
        make_leq_top_pairs(Ys,XYs).
make_leq_top_pairs([Y|Ys],[Y=<top|XYs]) :-
        make_leq_top_pairs(Ys,XYs).
*/
 

